#dimContentWrapper{
    margin-top:100px;
}

#sessions{
    margin-bottom:30px;
}

#nodes{
    margin-bottom:30px;
}

.bootbox-close-button{
  display: none;
}

#graphNameWrapper{
    position: absolute;
    top: 56px;
    display: block;
    width: 100%;
    height:auto;
    background-color: #002349;
    padding: 0px 20px;
    position: fixed;
    z-index: 101;
}

#graphNameWrapper li{
color: rgb(219, 219, 219);
display: inline-block;
}

#graphName{
margin: 0;
}

.navbar{
background: rgb(0,64,133);
background: linear-gradient(90deg, rgba(0,64,133,1) 0%, rgba(0,52,107,1) 20%, rgba(0,35,73,1) 100%);
margin-bottom:100px;
}

.navbar-brand{
  margin-left: 20px;
}

.navbar-brand span{
margin-left: 10px;
}

.navbar-nav.navbar-center {
position: absolute;
left: 50%;
transform: translatex(-50%);
}

.navbar-nav.navbar-center div {
margin-right:10px;
}

.navbar-nav button {
background-color: transparent;
border:1px white solid;
}

.navbar-nav .btn.btn-secondary.active{
background-color: white;
border:1px white solid;
color:rgba(0,35,73,1);
}

.navbar-nav .btn.btn-secondary:hover{
background-color: rgb(216, 216, 216);
border:1px rgb(216, 216, 216) solid;
color:rgba(0,35,73,1);
}

.btn.btn.btn-secondary{
    background-color: #0059a5;
}

.btn.btn.btn-secondary:hover{
    background-color: rgba(0,35,73,1);
}

.inactiveLink {
    pointer-events: none;
    cursor: default;
 }

 tr td:nth-child(n+2),tr th:nth-child(n+2) { 
  text-align: right;
}

tr td.actions button{
  padding: 1px 4px;
  font-size:18px;
  background-color: transparent !important;
  border: none;
  color: #0d6efd;
  display: inline-block;
}

tr td.actions button:hover{
  color: #092650;
}

tr td.actions button:disabled{
  color: #535353;
}

tr td.details a:hover{
  color: #092650;
}

.status svg {
    width: 100%;
    height: 20px;
    position: relative;
  }
  
.status {
  width: 150px;
  height:30px;
  overflow: hidden;
}


.tooltip-inner h5 {
  font-size: 1rem;
  font-weight: 600;
  margin: 0px;
}

.tooltip-inner p {
  font-size: .85rem;
  font-weight: 300;
  margin-bottom: 10px;
}

.tooltip-inner{
  transition: 0s visibility;
  visibility: visible;
}

.table-container {
    max-width: 75%;
    overflow-x: auto;
    margin: 0 auto;
}

.table {
    width: 100%;
    table-layout: auto; /* Let the columns resize based on content */
}

.table th, .table td {
    max-width: 800px; /* Limit maximum width of each column */
}

.wrap-text {
  white-space: normal !important;
  word-break: break-word;
}