# THIS FILE IS GENERATED BY SETUP.PY
# DO NOT MODIFY BY HAND
version = '6.0.0'
git_version = 'd266e83f3d79f1413c264ae387c41e81097edbc4'
full_version = '6.0.0'
is_release = True

if not is_release:
    version = full_version